/* USER CODE BEGIN Header */
/**
  ******************************************************************************
  * @file           : main.c
  * @brief          : Main program body
  ******************************************************************************
  * @attention
  *
  * Copyright (c) 2022 STMicroelectronics.
  * All rights reserved.
  *
  * This software is licensed under terms that can be found in the LICENSE file
  * in the root directory of this software component.
  * If no LICENSE file comes with this software, it is provided AS-IS.
  *
  ******************************************************************************
  */
/* USER CODE END Header */
/* Includes ------------------------------------------------------------------*/
#include "main.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */

/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */

/* USER CODE END PTD */

/* Private define ------------------------------------------------------------*/
/* USER CODE BEGIN PD */
/* USER CODE END PD */

/* Private macro -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/

TIM_HandleTypeDef htim3;

/* USER CODE BEGIN PV */
/*
	Task table.
	In our simple scheduler, the task table contains the fixed number of tasks

	Pa3cio Bulic, 26.10.2021
*/
task_table_t task_table[MAX_TASKS];

/*
	Task stacks.
	In our simple scheduler, each task has 1KB stack.
	The stacks are put in a linear RAM region of size MAX_TASKS x 1KB.

	Pa3cio Bulic, 26.10.2021
*/
uint32_t tasks_stack_region[MAX_TASKS * TASKS_STACK_SIZE];

// this is the address where all tasks stacks starts:
void *pTasksStackBase =  (void *)(&tasks_stack_region[MAX_TASKS * TASKS_STACK_SIZE]);

int current_task = -1;
/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
void SystemClock_Config(void);
static void MX_GPIO_Init(void);
static void MX_TIM3_Init(void);
/* USER CODE BEGIN PFP */

/* USER CODE END PFP */

/* Private user code ---------------------------------------------------------*/
/* USER CODE BEGIN 0 */


/*
	The function reads PSP and returns its value in result.
 	This function should be defined into the same translational module in which is used
 	as it is defined as static inline!

 	Pa3cio Bulic, 26.10.2021
*/
static __attribute__((always_inline)) void* read_process_stack_pointer(void) {
	void* result;
	/* read special register PSP into a general-purpose register (picked by compiler)
	  and write it to result: */
	asm volatile ( "MRS %0, PSP\n\t"
					: "=r" (result) );

	return (result);
}

/*
	The function writes the ptr to PSP.
 	This function should be defined into the same translational module in which is used
 	as it is defined as static inline!

 	Pa3cio Bulic, 26.10.2021
*/
static __attribute__((always_inline)) void write_process_stack_pointer(void* ptr) {

	asm volatile ( "MSR PSP, %0\n\t"
					:
					: "r" (ptr) );
}

/*
	The is the context switcher.

	1. save psp to sp of the current taks (in task table entry)
	2. select new task in a round-robin fashion (cyclic increment of the task index)
	3. write the sp of the new task into psp
	4. that's all :)

 	Pa3cio Bulic, 26.10.2021
*/
void switch_context (void) {
    if (current_task != -1) { // ali sem prekinil main()? Preskoci.
    	task_table[current_task].sp = read_process_stack_pointer();
    }

	// select a new task in a round-robin fashion:
	current_task++;
	if (current_task == MAX_TASKS) {
          current_task = 0;
	}

	// select a new psp:
	write_process_stack_pointer( task_table[current_task].sp );
}


/* USER CODE END 0 */

/**
  * @brief  The application entry point.
  * @retval int
  */
int main(void)
{
  /* USER CODE BEGIN 1 */

  /* USER CODE END 1 */

  /* MCU Configuration--------------------------------------------------------*/

  /* Reset of all peripherals, Initializes the Flash interface and the Systick. */
  HAL_Init();

  /* USER CODE BEGIN Init */

  /* USER CODE END Init */

  /* Configure the system clock */
  SystemClock_Config();

  /* USER CODE BEGIN SysInit */

  /* USER CODE END SysInit */

  /* Initialize all configured peripherals */
  MX_GPIO_Init();
  MX_TIM3_Init();
  /* USER CODE BEGIN 2 */
  HAL_GPIO_WritePin(GPIOJ, LED_RED_Pin, GPIO_PIN_RESET);
  HAL_GPIO_WritePin(GPIOJ, LED_GREEN_Pin, GPIO_PIN_RESET);


  init_tasks();

  // Start timer:
  if (HAL_TIM_Base_Start_IT(&htim3) != HAL_OK)
  {
	  Error_Handler();
  }
  /* USER CODE END 2 */

  /* Infinite loop */
  /* USER CODE BEGIN WHILE */
  while (1)
  {
    /* USER CODE END WHILE */

    /* USER CODE BEGIN 3 */
  }
  /* USER CODE END 3 */
}

/**
  * @brief System Clock Configuration
  * @retval None
  */
void SystemClock_Config(void)
{
  RCC_OscInitTypeDef RCC_OscInitStruct = {0};
  RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};

  /** Configure the main internal regulator output voltage
  */
  __HAL_RCC_PWR_CLK_ENABLE();
  __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE3);

  /** Initializes the RCC Oscillators according to the specified parameters
  * in the RCC_OscInitTypeDef structure.
  */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSI;
  RCC_OscInitStruct.HSIState = RCC_HSI_ON;
  RCC_OscInitStruct.HSICalibrationValue = RCC_HSICALIBRATION_DEFAULT;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_NONE;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    Error_Handler();
  }

  /** Initializes the CPU, AHB and APB buses clocks
  */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_HSI;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV1;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_0) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
  * @brief TIM3 Initialization Function
  * @param None
  * @retval None
  */
static void MX_TIM3_Init(void)
{

  /* USER CODE BEGIN TIM3_Init 0 */

  /* USER CODE END TIM3_Init 0 */

  TIM_ClockConfigTypeDef sClockSourceConfig = {0};
  TIM_MasterConfigTypeDef sMasterConfig = {0};

  /* USER CODE BEGIN TIM3_Init 1 */

  /* USER CODE END TIM3_Init 1 */
  htim3.Instance = TIM3;
  htim3.Init.Prescaler = PRESCALER;
  htim3.Init.CounterMode = TIM_COUNTERMODE_UP;
  htim3.Init.Period = PERIOD;
  htim3.Init.ClockDivision = TIM_CLOCKDIVISION_DIV1;
  htim3.Init.AutoReloadPreload = TIM_AUTORELOAD_PRELOAD_DISABLE;
  if (HAL_TIM_Base_Init(&htim3) != HAL_OK)
  {
    Error_Handler();
  }
  sClockSourceConfig.ClockSource = TIM_CLOCKSOURCE_INTERNAL;
  if (HAL_TIM_ConfigClockSource(&htim3, &sClockSourceConfig) != HAL_OK)
  {
    Error_Handler();
  }
  sMasterConfig.MasterOutputTrigger = TIM_TRGO_RESET;
  sMasterConfig.MasterSlaveMode = TIM_MASTERSLAVEMODE_DISABLE;
  if (HAL_TIMEx_MasterConfigSynchronization(&htim3, &sMasterConfig) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN TIM3_Init 2 */

  /* USER CODE END TIM3_Init 2 */

}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};

  /* GPIO Ports Clock Enable */
  __HAL_RCC_GPIOJ_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOJ, LED_RED_Pin|LED_GREEN_Pin, GPIO_PIN_SET);

  /*Configure GPIO pins : LED_RED_Pin LED_GREEN_Pin */
  GPIO_InitStruct.Pin = LED_RED_Pin|LED_GREEN_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_LOW;
  HAL_GPIO_Init(GPIOJ, &GPIO_InitStruct);

  /*Configure GPIO pin : USR_BTN_Pin */
  GPIO_InitStruct.Pin = USR_BTN_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_IT_RISING;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  HAL_GPIO_Init(USR_BTN_GPIO_Port, &GPIO_InitStruct);

  /* EXTI interrupt init*/
  HAL_NVIC_SetPriority(EXTI0_IRQn, 0, 0);
  HAL_NVIC_EnableIRQ(EXTI0_IRQn);

}

/* USER CODE BEGIN 4 */


/*
	The function creates a new task. Creating a new task involves:

	1. making space for the task's stake. Each task gets 4kB for the stack
	2. populating the initial HW stack frame for the task. HW stack frame initially resides
	   at the bottom of the task's stack. The HW stack frame is populated as follows:
	   PSR = 0x21000000 - this is the default value in the state register
	   PC = address of the task
	   LR = the address of the delete_task() function (if needed).
	   r12, r3-r0 = 0x00000000 - we may also pass the arguments into the tas via r0-r3, but this is not
	   	the case in our simple RTOS
	3. makes room for the initial uninitialized SW stack frame.
	4. sets the tasks stack pointer in the task table to point at the top of the SW stack frame

	After this steps a new task is ready to be executed for the first time when task switch occurs and
	the task is selected for execution.

 	Pa3cio Bulic, 26.10.2021
*/

void create_task (int task_id, void* task_stack_base_address, void (*pTask)(void)) {

	hw_stack_frame_t* ptask_hw_frame;
	sw_stack_frame_t* ptask_sw_frame;

	/* set the start address of the HW frame structure */
	ptask_hw_frame = (hw_stack_frame_t *) ((uint8_t *)task_stack_base_address
			- sizeof(hw_stack_frame_t));
	/* set the start address of the SW frame structure */
	ptask_sw_frame = (sw_stack_frame_t *) ((uint8_t *)task_stack_base_address
			- sizeof(hw_stack_frame_t) - sizeof(sw_stack_frame_t));

	/* populate the HW stack frame with initial values : */
	/* NOTE: HW stack for Task0 (main()) is ignored as main() uses MAIN STACK */
	ptask_hw_frame -> r0 = 0;
	ptask_hw_frame -> r1 = 0;
	ptask_hw_frame -> r2 = 0;
	ptask_hw_frame -> r3 = 0;
	ptask_hw_frame -> r12 = 0;
	ptask_hw_frame -> lr = 0; // in our simple RTOS the tasks never finish so there is no need to delete the task
	ptask_hw_frame -> pc = (uint32_t)pTask ;
	ptask_hw_frame -> xpsr = 0x01000000; // Set T bit (bit 24) in EPSR. The Cortex-M4 processor only
										 // supports execution of instructions in Thumb state.
										 // Attempting to execute instructions when the T bit is 0 (Debug state) results in a fault.
	ptask_sw_frame -> r4 = 4;
	ptask_sw_frame -> r5 = 5;
	ptask_sw_frame -> r6 = 6;
	ptask_sw_frame -> r7 = 7;
	ptask_sw_frame -> r8 = 8;
	ptask_sw_frame -> r9 = 9;
	ptask_sw_frame -> r10 = 10;
	ptask_sw_frame -> r11 = 11;

	// set the task's stack pointer:
	task_table[task_id].sp = (void *) ptask_sw_frame;
}

/*
	The function initializes and creates all tasks in the RTOS.
	Our RTOS is very simple and has fixed number of tasks that are created
	at the start of RTOS.
	The tasks never stop and the number of the running task never changes.

 	Pa3cio Bulic, 26.10.2021
*/

void init_tasks (void) {

	int i;
	void* task_stack_base_address;

     /* Task schedule
	    sets tasks in the schedule: */
    task_table[0].pTask = &Opravilo0;
	task_table[1].pTask = &Opravilo1;
	task_table[2].pTask = &Opravilo2;
	task_table[3].pTask = &Opravilo3;



	/* Create tasks....*/
	for (i = 0; i < MAX_TASKS; i++) {

		task_stack_base_address = (uint8_t *)pTasksStackBase - i*sizeof(uint32_t)*TASKS_STACK_SIZE;
		create_task (i, task_stack_base_address, task_table[i].pTask);
	}

    /* PSP has not been set until now, so set PSP to point to
	   the top of stack of the first interrupted task (Task 0): */
	write_process_stack_pointer(task_table[0].sp);
}



// Opravila:
void Opravilo0 (void) {
  while(1) {
	  HAL_GPIO_TogglePin(GPIOJ, LED_GREEN_Pin);
	  HAL_GPIO_WritePin(GPIOJ, LED_RED_Pin, GPIO_PIN_RESET);
	  HAL_Delay(100);
  }
}

void Opravilo1 (void) {
  while(1) {
	  HAL_GPIO_TogglePin(GPIOJ, LED_RED_Pin);
	  HAL_GPIO_WritePin(GPIOJ, LED_GREEN_Pin, GPIO_PIN_RESET);
	  HAL_Delay(100);
  }
}

void Opravilo2 (void) {
  while(1) {
	  HAL_GPIO_TogglePin(GPIOJ, LED_GREEN_Pin);
	  HAL_GPIO_WritePin(GPIOJ, LED_RED_Pin, GPIO_PIN_RESET);
	  HAL_Delay(300);
  }
}

void Opravilo3 (void) {
  while(1) {
	  HAL_GPIO_TogglePin(GPIOJ, LED_RED_Pin);
	  HAL_GPIO_WritePin(GPIOJ, LED_GREEN_Pin, GPIO_PIN_RESET);
	  HAL_Delay(300);
  }
}


HAL_TIM_PeriodElapsedCallback(TIM_HandleTypeDef *htim){
	if (htim == &htim3){
		__HAL_TIM_ENABLE(&htim3) ;
	}

}

/* USER CODE END 4 */

/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
void Error_Handler(void)
{
  /* USER CODE BEGIN Error_Handler_Debug */
  /* User can add his own implementation to report the HAL error return state */
  __disable_irq();
  while (1)
  {
  }
  /* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{
  /* USER CODE BEGIN 6 */
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */
